/*********************************************************************
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2022, KR Soft s.r.o.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the KR2013ApS nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/
#ifndef ROBOT_CONTROLLER_H
#define ROBOT_CONTROLLER_H

#include <climits>

#include "kr2/kord/protocol/DataDescriptions/Requests/ControlCommandStatus.h"
#include "kr2/kord/protocol/DataDescriptions/Requests/ControlCommandItems.h"
#include "kr2/kord/system/SystemEvent.h"

#include <boost/circular_buffer.hpp>

namespace kr2 {
namespace kord {

namespace kkpt = kr2::kord::protocol;

//
// CPU States
//
/// @brief Number of states available.
constexpr size_t num_cpu_states{5};

struct CPUStateData{
    CPUStateData():
        value_(0.0),
        id_(0)
    {
    }
    uint8_t id_;
    double value_;
};

struct CommandStatus{
        uint64_t token{};
        int8_t error_code{};
};


struct PendingStatus{
        uint64_t token{};
        unsigned int rc_token{};
        int times_checked{};
        bool composed{}; // composed if it is dependent on other statuses, like in Brakes (we have 7 statuses for each joint)
        int compos_left{}; // how many dependent results left
        int compos_total{}; // total number of dependencies
};

    /// @brief StatusBuffer structure for handling received CommandStatus
    ///
    template <typename T>
    struct StatusBuffer {

    public:
        boost::circular_buffer<T> statuses_{MAX_COMMAND_STATUSES};

        /**
         * @brief Add a Status
         *
         * \p a_status: T
        */
        void addStatus(const T& a_status)
        {
            //if (!isTokenPresent(a_status.token)){
               this->statuses_.push_back(a_status); 
            //}
        }


        /**
         * @brief Get the most recent status (and delete it from buffer)
         *
         * \p a_status:  T token
        */
        bool getRecentStatusAndDelete(T& a_status) {
            if (!statuses_.empty()){
                a_status = statuses_.front();
                statuses_.pop_front();
                return true;
            }
            else{
                return false;
            }
        }
        
        /**
         * @brief Reset the list of Statuses
         *
        */
        void reset()
        {
            this->statuses_.clear();
        }

        /**
         * @brief Check if the Statuses are empty
         * 
         * @return true if the list is empty
         *
        */
       bool is_empty() {
            return this->statuses_.empty();
       }

        private:
            static const unsigned int MAX_COMMAND_STATUSES{100};

            // bool isTokenPresent(const uint64_t& a_token){
            //     return std::find_if(statuses_.begin(), statuses_.end(), [a_token](CommandStatus i) {return i.token == a_token;}) != std::end(statuses_);
            // }
    };


class RobotController {
public:
    RobotController(){
        reset();
    }

    void reset() {
        rc_safety_flags_ = UINT_MAX;
        rc_hw_flags_ = UINT_MAX;
        rc_button_flags_ = UINT_MAX;
        master_speed_ = -1;
        safety_mode_ = UINT_MAX;
        motion_flags_ = UINT_MAX;
        req_system_id_ = kkpt::EControlCommandItems::eInvalid;
        req_sys_tx = 0;
        request_system_status_ = kkpt::EControlCommandStatus::eIdle;
    }

    // CPU state
    std::array<CPUStateData, num_cpu_states> cpu_state_;

    unsigned int rc_safety_flags_;
    unsigned int rc_hw_flags_;
    unsigned int rc_button_flags_;
    unsigned int rc_system_alarm_state_;
    double master_speed_;
    unsigned int safety_mode_;
    unsigned int motion_flags_;

    kkpt::EControlCommandItems req_system_id_ = kkpt::EControlCommandItems::eInvalid;
    int64_t req_sys_tx = 0;
    kkpt::EControlCommandStatus request_system_status_ = kkpt::EControlCommandStatus::eIdle;
    unsigned int request_error_{0};

    kkpt::SystemEvents system_events_;
};

};
};


#endif // ROBOT_CONTROLLER_H
