/*////////////////////////////////////////////////////////////////////////////
//
// (C) Copyright 2022 by Kassow Robots, ApS
//
// The information contained herein is confidential, proprietary to Kassow Robots,
// ApS, and considered a trade secret as defined in section 263 and section 264
// under the Danish Criminal Code. Use of this information by anyone
// other than authorized employees of Kassow Robots, ApS is granted only under a
// written non-disclosure agreement, expressly prescribing the scope and
// manner of such use.
//
// Authors: Ondrej Bruna  obr@kassowrobots.com
//
/////////////////////////////////////////////////////////////////////////////*/

#ifndef KR2_CBUN_KORD_ROBOT_IO_H
#define KR2_CBUN_KORD_ROBOT_IO_H

#pragma once

#include <cstdint>

#include <cmath>
#include <vector>
#include <array>

#include <kr2_rc_api/api_v1/io/iodata.h>

namespace kr2 {
namespace kord {

//
/// IO Board
//
/// @brief IO board current 4-20mA inputs.
constexpr size_t num_inputs_current_iob{2};

/// @brief IO board voltage 0-10V inputs.
constexpr size_t num_inputs_voltage_iob{2};

/// @brief IO board digital 0-30V inputs.
constexpr size_t num_inputs_digital_iob{16};

/// @brief IO board safe inputs. Performance level D cat 3.
///        Only activated by DIP switches, inactive otherwise.
constexpr size_t num_inputs_digital_safe_iob{4};


/// @brief IO board releays (NC/NO).
constexpr size_t num_outputs_relay_iob{4};

/// @brief IO board digital outputs.
constexpr size_t num_outputs_digital_iob{8};

/// @brief Safe outputs state control only if enabled from tablet.
constexpr size_t num_outputs_digital_safe_iob{4};

/// @brief IO board analog current output.
constexpr size_t num_outputs_current_iob{2};

/// @brief IO board voltage 0-10V output.
constexpr size_t num_outputs_voltage_iob{2};


//
// Tool Board
//
/// @brief Tool IO analog current input 4-20mA.
constexpr size_t num_inputs_current_tio{1};

/// @brief Tool IO analog voltage input 0-10V.
constexpr size_t num_inputs_voltage_tio{2};

/// @brief Tool IO digital outputs are configurable (12/24V). Max current 100mA.
constexpr size_t num_outputs_digital_tio{4};

/// @brief Tool IO power supply. Configurable to 12/24V. Max current 700mA.
constexpr size_t num_outputs_psu_tio{2};

/// @brief Tool IO analog outputs are configurable.
///        It can either serve as voltage 0-10V or current 4-20mA.
constexpr size_t num_outputs_analog_tio{2};

constexpr size_t num_analog_inputs = num_inputs_current_iob 
                                   + num_inputs_voltage_iob
                                   + num_inputs_current_tio
                                   + num_inputs_voltage_tio;

constexpr size_t num_analog_outputs = num_outputs_analog_tio
                                    + num_outputs_current_iob
                                    + num_outputs_voltage_iob;

/// @brief Holds state inputs as reported by RC API. 
/// It concerns both io board and tool board. 
/// Digital and analog inputs and outputs are considered.
///
/// Not available:
/// a) control over the safe outputs
/// b) control over quadrature LPR and index
class RobotIO {
public:
    using IORegister = uint64_t;
    using Analog = kr2rc_api::IOData::GPIOFloat;
    using Digital = kr2rc_api::IOData::GPIOInt64;

    RobotIO();

    void reset();
    
    void setOutputDigital(const std::vector<Digital>&);
    void setOutputAnalog(const std::vector<Analog>&);

    void setInputDigital(const std::vector<Digital>&);
    void setInputAnalog(const std::vector<Analog>&);

    IORegister getOutputRegister() const;
    IORegister getInputRegister() const;

    std::array<double, num_analog_outputs> getOutputAnalog() const;
    std::array<double, num_analog_inputs> getInputAnalog() const;

private:
    // Set a bit value to a register with an offset.
    inline void registerUpdate(IORegister& io_reg, const Digital& kr2_gpio, unsigned int offset);

    // Set a single bit in a register.
    inline void registerSetBit(IORegister&, unsigned int);

    // Reset a signle bit in a register.
    inline void registerResetBit(IORegister&, unsigned int);

private:
    /// Register of digital inputs from both iob and tio.
    IORegister inputs_register_;
    /// Register of digital outputs from both iob and tio.
    IORegister outputs_register_;

    std::array<double, num_analog_inputs> analog_inputs_;
    std::array<double, num_analog_outputs> analog_outputs_;
};

}
}

#endif // KR2_CBUN_KORD_ROBOT_IO_H
